﻿/**
 * EH: For purposes of debugging and testing, I disabled the generation of the
 * Excel spreadsheet. You can turn it back on by defining (as true), EXCEL, in
 * this file as well as in Run.cs.
 */

//#define EXCEL Test

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.IO;

namespace CPF_experement
{
    /// <summary>
    /// This is the entry point of the application. 
    /// </summary>
    class Program
    {
        private static string RESULTS_FILE_NAME = "CopathResults_map_ost003d_r208_smooth.csv";
       // private static string RESULTS_FILE_NAME = "CopathResults_grid_32x32_r104_smooth.csv";
        private static bool onlyReadInstances = false;

        private static string DAPath = "\\Dragonage\\brc202d";

        /// <summary>
        /// Simplest run possible with a randomly generated problem instance.
        /// </summary>
        public void SimpleRun()
        {
            Run runner = new Run();
            runner.openResultFile(RESULTS_FILE_NAME);
            runner.printResultsFileHeader();
            ProblemInstance instance = runner.generateProblemInstance(10, 3, 10);
            instance.Export("Test.instance");
            runner.solveGivenProblem(instance);            
            runner.closeResultsFile();
        }

        /// <summary>
        /// Runs a single instance, imported from a given filename.
        /// </summary>
        /// <param name="fileName"></param>
        public void RunInstance(string fileName)
        {
            Run runner = new Run();
            runner.openResultFile(RESULTS_FILE_NAME);
            runner.printResultsFileHeader();
            ProblemInstance instance = ProblemInstance.Import(Directory.GetCurrentDirectory() + "\\Instances\\" + fileName);
            runner.solveGivenProblem(instance);
            runner.closeResultsFile();
        }

        /// <summary>
        /// Runs a set of experiments.
        /// This function will generate a random instance (or load it from a file if it was already generated)
        /// </summary>
        public void RunExperimentSet(int[] gridSizes, int[] agentListSizes, int[] obstaclesProbs, int instances)
        {
            ProblemInstance instance;
            string instanceName;
            Run runner = new Run();
            runner.openResultFile(RESULTS_FILE_NAME);
            TextWriter output;
            

            bool continueFromLastRun = true; 
            string[] lineParts = null;
            if (File.Exists(Directory.GetCurrentDirectory() + "\\Instances\\current problem")) //if were continuing running from last time
            {
                TextReader input = new StreamReader(Directory.GetCurrentDirectory() + "\\Instances\\current problem");
                lineParts = input.ReadLine().Split(',');  //get the last problem
                input.Close();
                //runner.setOutOfTimeCounter();
            }
            else
            {
                runner.printResultsFileHeader();
                continueFromLastRun = false;
            }

            for (int gs = 0; gs < gridSizes.Length; gs++)
            {
                for (int obs = 0; obs < obstaclesProbs.Length; obs++)
                {
                    runner.resetOutOfTimeCounter();
                    //runner.setOutOfTimeCounter();
                    for (int ag = 0; ag < agentListSizes.Length; ag++)
                    {
                        if (gridSizes[gs] * gridSizes[gs] * (1-obstaclesProbs[obs]/100) - agentListSizes[ag] < 0)
                            continue;
                        for (int i = 0; i < instances; i++)
                        {
                            if (continueFromLastRun)  //set the latest problem
                            {
                                gs = int.Parse(lineParts[0]);
                                obs = int.Parse(lineParts[1]);
                                ag = int.Parse(lineParts[2]);
                                i = int.Parse(lineParts[3]);
                                for (int j = 4; j < lineParts.Length; j++)
                                {
                                    runner.outOfTimeCounter[j - 4] = int.Parse(lineParts[j]);
                                }
                                continueFromLastRun = false;
                                continue;
                            }
                            if ( runner.outOfTimeCounter.Length != 0 && runner.outOfTimeCounter.Sum() == runner.outOfTimeCounter.Length * Constants.MAX_FAIL_COUNT)
                                continue;
                            instanceName = "Instance-" + gridSizes[gs] + "-" + obstaclesProbs[obs] + "-" + agentListSizes[ag] + "-" + i;
                            try
                            {
                                instance = ProblemInstance.Import(Directory.GetCurrentDirectory() + "\\Instances\\"+instanceName);
                                instance.instanceId = i;
                            }
                            catch (Exception importException)
                            {
                                if (onlyReadInstances)
                                {
                                    Console.WriteLine("File " + instanceName + "  Dosent exists ");
                                    return;

                                    //Console.WriteLine("File Not Found!!!!");
                                    //Console.ReadLine();
                                }

                                instance = runner.generateProblemInstance(gridSizes[gs], agentListSizes[ag], obstaclesProbs[obs] * gridSizes[gs] * gridSizes[gs] / 100);
                                instance.ComputeSingleAgentShortestPaths();
                                instance.instanceId = i;
                                instance.Export("Instance-" + gridSizes[gs] + "-" + obstaclesProbs[obs] + "-" + agentListSizes[ag] + "-" + i);
                            }

                            runner.solveGivenProblem(instance);

                            //save the latest problem
                            if (File.Exists(Directory.GetCurrentDirectory() + "\\Instances\\current problem"))
                                File.Delete(Directory.GetCurrentDirectory() + "\\Instances\\current problem");
                            output = new StreamWriter(Directory.GetCurrentDirectory() + "\\Instances\\current problem");
                            output.Write("{0},{1},{2},{3}", gs, obs, ag, i);
                            for (int j = 0; j < runner.outOfTimeCounter.Length; j++)
                            {
                                output.Write("," + runner.outOfTimeCounter[j]);
                            }
                            output.Close();
                            //Console.ReadLine();
                        }
                    }
                }
            }
            runner.closeResultsFile();                    
        }
       
        
        /// <summary>
        /// dragon age experiment
        /// </summary>
        /// <param name="gridSizes"></param>
        /// <param name="agentListSizes"></param>
        /// <param name="obstaclesProbs"></param>
        /// <param name="instances"></param>
        public void RunDragonAgeExperimentSet( int instances)
        {
            ProblemInstance instance;
            string instanceName;
            Run runner = new Run();
            runner.openResultFile(RESULTS_FILE_NAME);
            TextWriter output;
            //int[] agentListSizes = { 5, 10, 15, 20, 25, 30, 40, 50, 70, 100, 130, 140, 170, 200, 250, 300 };
            int[] agentListSizes = { 150 };
            //int[] agentListSizes = { 70, 100 };

            bool continueFromLastRun = true;
            string[] lineParts = null;
            if (File.Exists(Directory.GetCurrentDirectory() + DAPath + "\\current problem")) //if were continuing running from last time
            {
                TextReader input = new StreamReader(Directory.GetCurrentDirectory() + DAPath+"\\current problem");
                lineParts = input.ReadLine().Split(',');  //get the last problem
                input.Close();
            }
            else
            {
                runner.printResultsFileHeader();
                continueFromLastRun = false;
            }

                    for (int ag = 0; ag < agentListSizes.Length; ag++)
                    {
                        for (int i = 0; i < instances; i++)
                        {
                            if (continueFromLastRun)  //set the latest problem
                            {
                                ag = int.Parse(lineParts[0]);
                                i = int.Parse(lineParts[1]);
                                for (int j = 2; j < lineParts.Length; j++)
                                {
                                    runner.outOfTimeCounter[j - 2] = int.Parse(lineParts[j]);
                                }
                                continueFromLastRun = false;
                                continue;
                            }
                            if (runner.outOfTimeCounter.Sum() == runner.outOfTimeCounter.Length * 20)
                                continue;
                            instanceName = agentListSizes[ag]+" agents i-"+i;
                            try
                            {
                                instance = ProblemInstance.Import(Directory.GetCurrentDirectory() + DAPath + "\\" + instanceName);
                            }
                            catch (Exception importException)
                            {
                                throw new Exception("missing map " + instanceName);
                            }

                            runner.solveGivenProblem(instance);

                            //save the latest problem
                            File.Delete(Directory.GetCurrentDirectory() + DAPath + "\\current problem");
                            output = new StreamWriter(Directory.GetCurrentDirectory() + DAPath + "\\current problem");
                            output.WriteLine("{0},{1}",ag, i);
                            for (int j = 0; j < runner.outOfTimeCounter.Length; j++)
                            {
                                output.Write("," + runner.outOfTimeCounter[j]);
                            }
                            output.Close();
                        }
                    }
            runner.closeResultsFile();
        }

        /// <summary>
        /// This is the starting point of the program. 
        /// Please do not commit this file due to changes in this method, 
        /// as it will probably be different for every one us.
        /// </summary>
        /// 
        public void Run_Grids(int[] robots, int grid_size)
        {
            int[] seeds = new int [] { 3455, 12, 45, 78, 8546, 474, 188, 7899, 9, 774 };

            for (int r = 0; r < robots.Length; ++r)
            {
                for (int sd = 0; sd < 10; ++sd)
                {
                    string filename = "grid_" + grid_size + "x" + grid_size + "_r" + robots[r] + "_" + seeds[sd] + ".bgu";
                    Console.WriteLine("Instance grid");
                    Console.WriteLine("  Size:" + grid_size);
                    Console.WriteLine("  Robots:" + robots[r]);
                    Console.WriteLine("  Seed:" + seeds[sd]);
                    Console.WriteLine(filename);
                    RunInstance("ijcai2017\\grid\\" + filename);
                  }
            }
        }

        public void Run_Grid_Smooth(int robots, int grid_size)
        {
            int[] seeds = new int[] { 3455, 12, 45, 78, 8546, 474, 188, 7899, 9, 774 };

            for (int sd = 0; sd < 10; ++sd)
                {
                    string filename = "grid_" + grid_size + "x" + grid_size + "_r" + robots + "_" + seeds[sd] + ".bgu";
                    Console.WriteLine("Instance grid");
                    Console.WriteLine("  Size:" + grid_size);
                    Console.WriteLine("  Robots:" + robots);
                    Console.WriteLine("  Seed:" + seeds[sd]);
                    Console.WriteLine(filename);
                    RunInstance("ijcai2017\\grid\\" + filename);
                }
        }

        public void Run_Maps(int[] robots, string map_name)
        {
            int[] seeds = new int[] { 3455, 12, 45, 78, 8546, 474, 188, 7899, 9, 774 };

            for (int r = 0; r < robots.Length; ++r)
            {
                for (int sd = 0; sd < 10; ++sd)
                {
                    string filename = "map_" + map_name + ".map_r" + robots[r] + "_" + seeds[sd] + ".bgu";
                    Console.WriteLine("Instance map");
                    Console.WriteLine("  Map   :" + map_name);
                    Console.WriteLine("  Robots:" + robots[r]);
                    Console.WriteLine("  Seed  :" + seeds[sd]);
                    Console.WriteLine(filename);
                    RunInstance("ijcai2017\\maps\\" + filename);
                }
            }
        }
    
        public void Run_Maps_Distances(int[] distances, int robots, string map_name)
        {
            int[] seeds = new int[] { 3455, 12, 45, 78, 8546, 474, 188, 7899, 9, 774 };

            for (int d = 0; d < distances.Length; ++d)
            {
                for (int sd = 0; sd < 10; ++sd)
                {
                    string filename = "map_" + map_name + ".map_r" + robots + "_d" + distances[d] + "_" + seeds[sd] + ".bgu";
                    Console.WriteLine("Instance map");
                    Console.WriteLine("  Map     :" + map_name);
                    Console.WriteLine("  Robots  :" + robots);
                    Console.WriteLine("  Distance:" + distances[d]);
                    Console.WriteLine("  Seed    :" + seeds[sd]);
                    Console.WriteLine(filename);
                    RunInstance("ijcai2017\\maps\\" + filename);
                }
            }
        }

        public void Run_Map_Smooth(int robots, string map_name)
        {
            int[] seeds = new int[] { 3455, 12, 45, 78, 8546, 474, 188, 7899, 9, 774 };

            for (int sd = 0; sd < 10; ++sd)
            {
                 string filename = "map_" + map_name + ".map_r" + robots + "_" + seeds[sd] + ".bgu";
                 Console.WriteLine("Instance map");
                 Console.WriteLine("  Map   :" + map_name);
                 Console.WriteLine("  Robots:" + robots);
                 Console.WriteLine("  Seed  :" + seeds[sd]);
                 Console.WriteLine(filename);
                 RunInstance("ijcai2017\\maps\\" + filename);
             }
        }

        static void Main(string[] args)
        {
            Program me = new Program();
            //Program.RESULTS_FILE_NAME = Process.GetCurrentProcess().ProcessName + ".csv";
            TextWriterTraceListener tr1 = new TextWriterTraceListener(System.Console.Out);
            Debug.Listeners.Add(tr1);
            //if (System.Diagnostics.Debugger.IsAttached)
            //{
            //    int maxTime = Constants.MAX_TIME;
            //    Constants.MAX_TIME = int.MaxValue;
            //    Console.WriteLine("Use Time Limit? <(ENTER)-no,(1)-yes>");
            //    if (Console.ReadLine() == "1")
            //        Constants.MAX_TIME = maxTime;
            //}

            if (Directory.Exists(Directory.GetCurrentDirectory() + "\\Instances") == false)
            {
                Directory.CreateDirectory(Directory.GetCurrentDirectory() + "\\Instances");
            }


            //Console.WriteLine("Only Pre-Made Instances? <(ENTER)-no,(1)-yes>");
            //if (Console.ReadLine() == "1")
            //{
            //    Program.onlyReadInstances = true;
            //    Console.WriteLine("You Choose Only Pre-Made Instances");
            //}
            //else
            //    Console.WriteLine("You Choose To Allow Creation Of New Instances");

           // Console.ReadLine();

            //Program.onlyReadInstances = true;

            int instances = 100;

            //int[] gridSizes = new int[] { 3 };
            //Constants.MAX_TIME = 2400000;
            //int[] agentListSizes = new int[] { 2, 3, 4, 5, 6, 7, 8 };

            //int[] gridSizes = new int[] { 8 };
            //int[] agentListSizes = new int[] { 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18 };
            //int[] agentListSizes = new int[] { 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 42, 44, 46, 48, 50 };

            //int[] gridSizes = new int[] { 32 };
           // int[] agentListSizes = new int[] { 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 105, 110, 115, 120, 125, 130, 135, 140, 145, 150 };
            //int[] agentListSizes = new int[] { 20 };

           // int[] agentListSizes = new int[] { 4 };
            //int[] obstaclesProbs = new int[] { 0, 5, 10, 15, 20, 25, 30, 35 };

           // int[] obstaclesProbs = new int[] { 0,10,20,30 };

            int[] gridSizes = new int[] { 5/*16*/};
            int[] agentListSizes = new int[] { 10 }; //{/*2,*/ 2,3,4,5,6, 7, 10, 15, 20, 25,30,35, 40, 50,60, 75, /*100 */};
            int[] obstaclesProbs = new int[] { 20/*, 10, 20, 40*/ };

            //int[] gridSizes = new int[] { 8/*16*/};
            //int[] agentListSizes = new int[] { 5,10,15,20,25,30,35,40, 45, 50 }; //{/*2,*/ 2,3,4,5,6, 7, 10, 15, 20, 25,30,35, 40, 50,60, 75, /*100 */};
            //int[] obstaclesProbs = new int[] { 0/*, 10, 20, 40*/ };

            //int[] gridSizes = new int[] { 32/*16*/};
            //int[] agentListSizes = new int[] { 80 }; //{/*2,*/ 2,3,4,5,6, 7, 10, 15, 20, 25,30,35, 40, 50,60, 75, /*100 */};
            //int[] obstaclesProbs = new int[] { 20/*, 10, 20, 40*/ };

            //me.RunExperimentSet(gridSizes, agentListSizes, obstaclesProbs, instances);

            int[] sizes = new int[] { 8, 16, 32 };

            int[] robots_08 = new int[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 18, 20, 22, 24, 26, 28, 30, 32 };
            int[] robots_16 = new int[] { 1, 2, 3, 4, 5, 6, 7, 8, 10, 12, 14, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 64, 72, 80, 88, 96, 104, 112, 120, 128 };
            int[] robots_32 = new int[] { 1, 2, 3, 4, 5, 6, 7, 8, 10, 12, 14, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 64, 72, 80, 88, 96, 104, 112, 120, 128, 144, 160, 176, 192, 208, 224, 240, 256 };
            //int[] robots_32 = new int[] { 240 };

            int[] robots_maps = new int[] { 1, 2, 3, 4, 5, 6, 7, 8, 10, 12, 14, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 64, 72, 80, 88, 96, 104, 112, 120, 128, 144, 160, 176, 192, 208, 224, 240, 256 };
            int[] distances = new int[] { 8, 16, 24, 32, 40, 48, 56, 64, 72, 80, 88, 96, 104, 112, 120, 128, 136, 144, 152, 160, 168, 176, 184, 192, 200, 208, 216, 224, 232, 240, 248, 256, 264, 272, 280, 288, 296, 304, 312, 320 };

            //me.Run_Grids(robots_08,  8);
            //me.Run_Grids(robots_16, 16);
            //me.Run_Grids(robots_32, 32);
            //me.Run_Grid_Smooth(104, 32);

            //me.Run_Maps(robots_maps, "den520d");
            //me.Run_Maps(robots_maps, "brc202d");
            //me.Run_Maps(robots_maps, "ost003d");

            //me.Run_Maps_Distances(distances, 16, "den520d");
            //me.Run_Maps_Distances(distances, 16, "brc202d");
            //me.Run_Maps_Distances(distances, 16, "ost003d");

            //me.Run_Maps_Distances(distances, 32, "den520d");
            //me.Run_Maps_Distances(distances, 32, "brc202d");
            //me.Run_Maps_Distances(distances, 32, "ost003d");
            me.Run_Map_Smooth(208, "ost003d");

            //int[] h_versions = new int[] { 2, 3, 4, 5 };

            //for (int i = 0; i < h_versions.Length; i++)
            //{
            //    if (File.Exists(Directory.GetCurrentDirectory() + "\\Instances\\current problem"))
            //        File.Delete(Directory.GetCurrentDirectory() + "\\Instances\\current problem");

            //    Constants.H_VER = h_versions[i];
            //    me.RunExperimentSet(gridSizes, agentListSizes, obstaclesProbs, instances);

            //}

            //me.RunDragonAgeExperimentSet(instances);

            // A function for running a single instance that is loaded from this file.
            // me.RunInstance("Nathan");
            // A function to be used by Eric's PDB code
            //me.runForPdb();
            Console.WriteLine("Press any key to continue...");
            Console.ReadLine();
        }    
    }
}
